;;theme.jl

;;Copyright (C) 2001  Martin Rudat <martin@whoever.com>

;; uses a base pixmap, and a bunch of masks to create all the decorations
;; used in the theme on the fly.

(defgroup Dark2 "Dark2")

(defgroup colors "colors" :group Dark2)

(defcustom Dark2:inactive-color (get-color "#808080")
  "color for inactive widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)
(defcustom Dark2:focused-color (get-color "#FFFFFF")
  "color for focused widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)
(defcustom Dark2:highlighted-color (get-color "#FFFF00")
  "color for highlighted widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)
(defcustom Dark2:clicked-color (get-color "#FFFF00")
  "color for clicked widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)
(defcustom Dark2:inactive-highlighted-color (get-color "#FFFFFF")
  "color for inactive-highlighted widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)
(defcustom Dark2:inactive-clicked-color (get-color "#FFFF00")
  "color for inactive-clicked widgets"
  :type color
  :group (Dark2 colors)
; :after-set function
)

(define tile-image  (make-image "Dark2.png"))
(define tile-height (cdr (image-dimensions tile-image)))
(define tile-width  (car (image-dimensions tile-image)))

(define widget-colors
  (lambda ()
    `(
       (inactive             . ,Dark2:inactive-color)
       (focused              . ,Dark2:focused-color)
       (highlighted          . ,Dark2:highlighted-color)
       (clicked              . ,Dark2:clicked-color)
       (inactive-highlighted . ,Dark2:inactive-highlighted-color)
       (inactive-clicked     . ,Dark2:inactive-clicked-color)
    )
  )
)


(define make-pixel
  (lambda (x)
    (if (colorp x)
      `(,@(color-rgb-8 x) 255)
      (unless (nth 1 x) ; is it grey-scale?
	(append (make-list 3 (car x)) (list 255))
	(if (nth 3 x) ; does it have alpha?
	  x
	  `(,@x 255)
	)
      )
    )
  )
)

(define recolor-image
  (lambda (dest mask colori)
    (let*
      (
	(color
;         (if
;	    (and
;	      (listp colori)
;	      (or
;	        (listp  (nth 0 colori))
;		(colorp (nth 0 colori))
;	      )
;	    )
;	    (mapcar (function make-pixel) colori)
            (list (make-pixel colori))
;         )
	)
	(m-mx (car (image-dimensions mask)))
	(m-my (cdr (image-dimensions mask)))

	(transform
	  (lambda (x y)
	    (let*
	      (
		(m-p
		  (image-ref mask
		    (% x        m-mx)
		    (% y        m-my)
		))
	      )
	      (if (= (nth 3 m-p) 0)
	        (list 0 1 0 255)
                (if
                  (nth (nth 0 m-p) color)
                  (nth (nth 0 m-p) color)
                  (nth 0 color)
                )
	      )
	    )
	  )
	)
      )
      (image-fill transform dest)
      ; blargh... what's the point of an alpha channel if you can't use
      ; it to set the shape with?
      (set-image-shape-color dest (get-color "#000100"))
    )
  )
)

(define recolor-images
  (lambda (mask colors width height)
    (mapcar
      (function
	(lambda (state)
	  (cons
	    state
	    (let
	      (
		(img (make-sized-image width height))
	      )
	      (image-put img 'tiled t)
	      (recolor-image img mask (cdr (assoc state colors)))
	      img
	    )
	  )
	)
      )
      (list
	'focused          'inactive
	'highlighted      'inactive-highlighted
	'clicked          'inactive-clicked
      )
    )
  )
)

(define image-copy-part
  (lambda (dest src sx sy dx dy w h)
    (composite-images dest (crop-image src sx sy w h) dx dy)
  )
)

(define image-copy-part-recurse
  (lambda (dest tile sx sy dx dy w h)
    (let
      (
	(img (make-sized-image w h))
      )
      (image-copy img tile sx sy)
      (composite-images dest img dx dy)
    )
  )
)

(define image-copy
  (lambda (dest tile ox oy)
    (let
      (
	(d-mx (car (image-dimensions dest)))
	(d-my (cdr (image-dimensions dest)))
	(t-mx (car (image-dimensions tile)))
	(t-my (cdr (image-dimensions tile)))
      )
      (let*
	(
	  (ox   (% ox t-mx))
	  (oy   (% oy t-mx))
	  (ox   (if (< ox 0) (+ ox t-mx) ox))
	  (oy   (if (< oy 0) (+ oy t-my) oy))
	)
	(if
	  (and
	    (<= (+ ox d-mx) t-mx)
	    (<= (+ oy d-my) t-my)
	  )
	  (image-copy-part dest tile ox oy 0 0 d-mx d-my)
	  (let
	    (
	      (left-width    (min (- t-mx ox) d-mx))
	      (top-height    (min (- t-my oy) d-my))
	    )
            (let
	      (
		(right-width   (- d-mx left-width))
		(bottom-height (- d-my top-height))
	      )
	      (if (and (> left-width  0) (> top-height    0))
		(image-copy-part-recurse dest tile  ox oy          0          0  left-width    top-height)
	      )
	      (if (and (> right-width 0) (> top-height    0))
		(image-copy-part-recurse dest tile   0 oy left-width          0  left-width    top-height)
	      )
	      (if (and (> left-width  0) (> bottom-height 0))
		(image-copy-part-recurse dest tile  ox  0          0 top-height  left-width bottom-height)
	      )
	      (if (and (> right-width 0) (> bottom-height 0))
		(image-copy-part-recurse dest tile   0  0 left-width top-height right-width bottom-height)
	      )
            )
	  )
	)
      )
    )
  )
)

(define render-image
  (lambda (state render width height)
    (let
      (
	(img (make-sized-image width height))
      )
      (image-put img 'tiled t)
      (render img state)
      img
    )
  )
)

(define render-images
  (lambda (render width height)
    (mapcar
      (function
	(lambda (state)
	  (cons
	    state
	    (render-image state render width height)
	  )
	)
      )
      (list
	'focused          'inactive
	'highlighted      'inactive-highlighted
	'clicked          'inactive-clicked
      )
    )
  )
)

(define set-pixels
  (lambda (img color pixels)
    (let
      (
        (color (make-pixel color))
      )
      (mapcar
	(function
	  (lambda (x)
	    (image-set img (car x) (cdr x) color)
	  )
	)
	pixels
      )
    )
  )
)

(define (frame-part w class)
  (let
    (
      (f (window-frame w))
    )
    (catch 'found
      (mapc
	(function
	  (lambda (x)
	    (if
	      (=
		(cdr (assoc 'class x))
		class
	      )
	      (throw 'found x)
	    )
	  )
	)
	f
      )
    )
  )
)

(define (search-windows class data)
  (catch 'found
    (mapc
      (lambda (w)
        (if
          (and
            (equal (window-get w 'current-frame-style ) 'Dark2 )
	    (equal (cdr (assoc '-data (frame-part w class))) data)
          )
          (throw 'found w)
        )
      )
      (managed-windows)
    )
  )
)

(define (cache-render w class renderer #!rest args)
  (let*
    (
      (f (frame-part w class))
      (images (assoc '-images f))
      (data   (assoc '-data f))
    )
    (unless (equal (cdr data) args)
      (let
	(
	  (w2 (search-windows class args))
	)
	(if w2
	  (rplacd images (cdr (assoc '-images (frame-part w2 class))))
	  (rplacd images (apply renderer args))
        )
        (rplacd data args)
      )
    )
    (cdr images)
  )
)

(define (window-height w)
  (if
    (window-get w 'shaded )
    0
    (cdr (window-dimensions w))
  )
)

(define (window-width w)
  (car (window-dimensions w))
)

(define (frame-height w)
  (+ (window-height w) 15)
)

;; title-bar...
(define menu-mask          (make-image "menu.png"))
(define menu-fg-images     (recolor-images menu-mask     (widget-colors) 15 15))
(define menu-part
 `(
    (class . menu-button)
    (background .
     ,(let*
	(
	  (img (make-sized-image 15 15))
	)
	(image-put img 'tiled t)
	(image-copy img tile-image 0 -15)
	img
      )
    )
    (foreground .
     ,(lambda (w)
	(if
	  (window-icon-image w)
	  (scale-image (window-icon-image w) 15 15)
	  menu-fg-images
	)
      )
    )
    (x-justify . center)
    (y-justify . center)
    (height . 15)
    (left-edge . 0)
    (top-edge . -15)
    (width . 15)
    (removable . t)
  )
)

(define (title-part-ox w)
  (- 15
    (apply +
      (mapcar
	(lambda (x)
	  (case x
	    ((menu-button)  15)
	    (t               0)
	  )
	)
	(window-get w 'removed-classes )
      )
    )
  )
)
(define title-part
 `(
    (class . title)
    (background .
     ,(lambda (w)
	(cache-render w 'title
	  (lambda (tile-image ox)
	    (let
	      (
		(img (make-sized-image tile-width 15))
	      )
	      (image-put img 'tiled t)
	      (image-copy img tile-image ox -15)
	      img
	    )
	  )
	  tile-image (title-part-ox w)
        )
      )
    )
    (foreground . ,(widget-colors))
    (height . 15)
    (left-edge . ,title-part-ox)
    (right-edge .
     ,(lambda (w)
        (- 44
	  (apply +
	    (mapcar
	      (lambda (x)
		(case x
		  ((iconify-button maximize-button)
		      15)
		  ((close-button)
		      14)
		  (t
		       0)
		)
	      )
	      (window-get w 'removed-classes )
	    )
	  )
        )
      )
    )
    (text . ,window-name)
    (top-edge . -15)
    (x-justify . 3)
    (y-justify . center)
  )
)

(define (iconify-part-ox w)
  (- (window-width w)
    (- 44
      (apply +
	(mapcar
	  (lambda (x)
	    (case x
	      ((maximize-button)
		  15)
	      ((close-button)
		  14)
	      (t
		   0)
	    )
	  )
	  (window-get w 'removed-classes )
	)
      )
    )
  )
)
(define iconify-mask       (make-image "iconify.png"))
(define iconify-part
 `(
    (class . iconify-button)
    (background .
     ,(lambda (w)
	(cache-render w 'iconify-button
	  (lambda (tile-image ox)
	    (let
	      (
		(img (make-sized-image 15 15))
	      )
	      (image-put img 'tiled t)
	      (image-copy img tile-image ox -15)
	      img
	    )
	  )
	  tile-image (iconify-part-ox w)
        )
      )
    )
    (foreground .
      ,(recolor-images iconify-mask  (widget-colors) 15 15)
    )
    (left-edge . ,iconify-part-ox)
    (x-justify . center)
    (y-justify . center)
    (top-edge . -15)
    (height . 15)
    (width . 15)
    (removable . t)
  )
)

(define (maximize-part-ox w)
  (- (window-width w)
    (- 29
      (apply +
	(mapcar
	  (lambda (x)
	    (case x
	      ((maximize-button)
		  15)
	      ((close-button)
		  14)
	      (t
		   0)
	    )
	  )
	  (window-get w 'removed-classes )
	)
      )
    )
  )
)
(define maximize-mask      (make-image "maximize.png"))
(define maximize-part
 `(
    (class . maximize-button)
    (background .
     ,(lambda (w)
	(cache-render w 'maximize-button
	  (lambda (tile-image ox)
	    (let
	      (
		(img (make-sized-image 15 15))
	      )
	      (image-put img 'tiled t)
	      (image-copy img tile-image ox -15)
	      img
	    )
	  )
	  tile-image (maximize-part-ox w)
        )
      )
    )
    (foreground .
     ,(recolor-images maximize-mask (widget-colors) 15 15)
    )
    (left-edge . ,maximize-part-ox)
    (x-justify . center)
    (y-justify . center)
    (top-edge . -15)
    (height . 15)
    (width . 15)
    (removable . t)
  )
)

(define close-mask         (make-image "close.png"))
(define close-part
 `(
    (class . close-button)
    (background .
     ,(lambda (w)
	(cache-render w 'close-button
	  (lambda (tile-image ox)
	    (let
	      (
		(img (make-sized-image 14 15))
	      )
	      (image-put img 'tiled t)
	      (image-copy img tile-image ox -15)
	      img
	    )
	  )
	  tile-image (- (window-width w) 14)
        )
      )
    )
    (foreground .
     ,(recolor-images close-mask    (widget-colors) 14 15)
    )
    (x-justify . center)
    (y-justify . center)
    (right-edge . 0)
    (top-edge . -15)
    (height . 15)
    (width . 14)
    (removable . t)
  )
)

;; edge parts...
(define center-left-part
 `(
    (class . left-border)
    (background .
     ,(lambda (w)
	(cache-render w 'left-border
          (lambda (tile-image)
	    (render-images
	      (lambda (img state)
		(image-copy img tile-image -4 -15)
		(composite-images
		  img
		  (make-sized-image
		    1 tile-height
		    (cdr (assoc state (widget-colors)))
		  )
		  0 0
		)
	      )
	      4 tile-height
	    )
          )
          tile-image
        )
      )
    )
    (left-edge . -4)
    (top-edge . -15)
    (height . ,frame-height)
    (width . 4)
  )
)

(define center-right-part
  `(
    (class . right-border)
    (background .
     ,(lambda (w)
	(cache-render w 'right-border
          (lambda (tile-image ox)
	    (render-images
	      (lambda (img state)
		(image-copy img tile-image ox -15)
		(composite-images
		  img
		  (make-sized-image
		    1 tile-height
		    (cdr (assoc state (widget-colors)))
		  )
		  3 0
		)
	      )
	      4 tile-height
	    )
          )
          tile-image (window-width w)
        )
      )
    )
    (right-edge . -4)
    (top-edge . -15)
    (height . ,frame-height)
    (width . 4)
  )
)

(define top-center-part
 `(
    (class . top-border)
    (background .
     ,(lambda (w)
	(cache-render w 'top-border
          (lambda (tile-image)
	    (render-images
	      (lambda (img state)
		(image-copy img tile-image 0 -19)
		(composite-images
		  img
		  (make-sized-image
		    tile-width 1
		    (cdr (assoc state (widget-colors)))
		  )
		  0 0
		)
	      )
	      tile-width 4
	    )
          )
          tile-image
        )
      )
    )
    (left-edge . 0)
    (right-edge . 0)
    (top-edge . -19)
    (height . 4)
  )
)

(define bottom-center-part
  `(
    (class . bottom-border)
    (background .
     ,(lambda (w)
	(cache-render w 'bottom-border
          (lambda (tile-image oy)
	    (render-images
	      (lambda (img state)
		(image-copy img tile-image 0 oy)
		(composite-images
		  img
		  (make-sized-image
		    tile-width 1
		    (cdr (assoc state (widget-colors)))
		  )
		  0 3
		)
	      )
	      tile-width 4
	    )
          )
          tile-image (window-height w)
        )
      )
    )
    (left-edge . 0)
    (right-edge . 0)
    ;(top-edge .  ; aargh! why doesn't this work?!
    ; ,(lambda (w)
    ;    (window-height w)
    ;  )
    ;)
    (bottom-edge .
     ,(lambda (w)
	(-
	  (if
	    (window-get w 'shaded )
	    (cdr (window-dimensions w))
	    0
	  )
	  4
	)
      )
    )
    (height . 4)
  )
)

;; corner parts...
(define (top-left-images tile-image)
  (render-images
    (lambda (img state)
      (image-copy
	img
	tile-image
	-4 -19
      )
      (set-pixels
	img
	(cdr (assoc state (widget-colors)))
	`(
				  (0 . 3)
			  (1 . 2)
		  (2 . 1)
	  (3 . 0)
	)
      )
      (set-pixels
	img
	(get-color "#000100")
	`(
	  (0 . 0) (0 . 1) (0 . 2)
	  (1 . 0) (1 . 1)
	  (2 . 0)

	)
      )
      (set-image-shape-color img (get-color "#000100"))
    )
    4 4
  )
)
(define top-left-part
 `(
    (class . top-left-corner)
    (background .
     ,(lambda (w)
	(cache-render w 'top-left-corner
          top-left-images
          tile-image
        )
      )
    )
    (left-edge . -4)
    (top-edge . -19)
    (height . 4)
    (width . 4)
  )
)

(define (top-right-images tile-image ox)
  (render-images
    (lambda (img state)
      (image-copy
	img
	tile-image
	ox -19
      )
      (set-pixels
	img
	(cdr (assoc state (widget-colors)))
	`(
	  (0 . 0)
		  (1 . 1)
			  (2 . 2)
				  (3 . 3)
	)
      )
      (set-pixels
	img
	(get-color "#000100")
	`(

	  (1 . 0)
	  (2 . 0) (2 . 1)
	  (3 . 0) (3 . 1) (3 . 2)
	)
      )
     (set-image-shape-color img (get-color "#000100"))
    )
    4 4
  )
)
(define top-right-part
 `(
    (class . top-right-corner)
    (background .
     ,(lambda (w)
	(cache-render w 'top-right-corner
          top-right-images
          tile-image (window-width w)
        )
      )
    )
    (right-edge . -4)
    (top-edge . -19)
    (height . 4)
    (width . 4)
  )
)

(define (bottom-left-images tile-image oy)
  (render-images
    (lambda (img state)
      (image-copy
	img
	tile-image
	-4 oy
      )
      (set-pixels
	img
	(cdr (assoc state (widget-colors)))
	`(
	  (0 . 0)
		  (1 . 1)
			  (2 . 2)
				  (3 . 3)
	)
      )
      (set-pixels
	img
	(get-color "#000100")
	`(
		  (0 . 1) (0 . 2) (0 . 3)
			  (1 . 2) (1 . 3)
				  (2 . 3)

	)
      )
     (set-image-shape-color img (get-color "#000100"))
    )
    4 4
  )
)
(define bottom-left-part
 `(
    (class . bottom-left-corner)
    (background .
     ,(lambda (w)
	(cache-render w 'bottom-left-corner
          bottom-left-images
          tile-image (window-height w)
        )
      )
    )
    (left-edge . -4)
    ;(top-edge .  ; aargh! why doesn't this work?!
    ; ,(lambda (w)
    ;    (window-height w)
    ;  )
    ;)
    (bottom-edge .
     ,(lambda (w)
	(-
	  (if
	    (window-get w 'shaded )
	    (cdr (window-dimensions w))
	    0
	  )
	  4
	)
      )
    )
    (height . 4)
    (width . 4)
  )
)

(define (bottom-right-images tile-image ox oy)
  (render-images
    (lambda (img state)
      (image-copy
	img
	tile-image
        ox oy
      )
      (set-pixels
	img
	(cdr (assoc state (widget-colors)))
	`(
				  (0 . 3)
			  (1 . 2)
		  (2 . 1)
	  (3 . 0)
	)
      )
      (set-pixels
	img
	(get-color "#000100")
	`(

				  (1 . 3)
			  (2 . 2) (2 . 3)
		  (3 . 1) (3 . 2) (3 . 3)
	)
      )
      (set-image-shape-color img (get-color "#000100"))
    )
    4 4
  )
)
(define bottom-right-part
 `(
    (class . bottom-right-corner)
    (background .
     ,(lambda (w)
	(cache-render w 'bottom-right-corner
          bottom-right-images
          tile-image (window-height w) (window-width w)
        )
      )
    )
    (right-edge . -4)
    ;(top-edge .  ; aargh! why doesn't this work?!
    ; ,(lambda (w)
    ;    (window-height w)
    ;  )
    ;)
    (bottom-edge .
     ,(lambda (w)
	(-
	  (if
	    (window-get w 'shaded )
	    (cdr (window-dimensions w))
	    0
	  )
	  4
	)
      )
    )
    (height . 4)
    (width . 4)
  )
)

(define (frame w)
  (mapcar
    (lambda (x)
      (append x
        `(
          (-data . ,(list ))
          (-images . ,(list ))
          (-window . ,w)
	)
      )
    )
    `(
      ;; title-bar...
      ,menu-part
      ,title-part
      ,iconify-part
      ,maximize-part
      ,close-part
      ;; frame...
      ,top-center-part
      ,center-left-part
      ,center-right-part
      ,bottom-center-part
      ;; corners...
      ,top-left-part
      ,top-right-part
      ,bottom-left-part
      ,bottom-right-part
    )
  )
)

(define (shaped-frame w)
  (mapcar
    (lambda (x)
      (append x
        `(
          (-data . ,(list ))
          (-images . ,(list ))
          (-window . ,w)
	)
      )
    )
    `(
      ;; title-bar...
      ,menu-part
      ,title-part
      ,iconify-part
      ,maximize-part
      ,close-part
    )
  )
)

(add-frame-style 'Dark2
  (lambda (w type)
    (case type
      ((shaped shaped-transient)
	(shaped-frame w)
      )
      ((shaded shaded-transient)
	(if (window-shaped-p w)
	  (shaped-frame w)
	  (frame w)
	)
      )
      ((unframed)
	nil-frame
      )
      (t ;default transient
	(frame w)
      )
    )
  )
)
